#include <foundation/memory/memory_globals.h>
#include <foundation/memory/arena_allocator.h>
#include <foundation/memory/heap_allocator.h>

#include <foundation/platform.h>
#include <foundation/collection/string.h>
#include <foundation/os/command_line.h>
#include <foundation/math/random.h>
#include <foundation/error/error_context.h>
#include <foundation/error/callstack.h>
#include <foundation/error/exception_handling_win.h>

#include <application/application.h>
#include <application/application_options.h>
#include <foundation/os/android/apk_asset_manager.h>
#include <foundation/os/android/apk_ext_manager.h>

#if defined(HAS_STEAM) && !defined(DEVELOPMENT)
	#include <steam/debug/minidump.h>
#endif

#include <game/game.h>

#include <jni.h>
#include <errno.h>

#include <EGL/egl.h>
#include <GLES/gl.h>

#include <foundation/error/logging.h>
#include <foundation/error/exception_handling_android.h>
#include <android/native_activity.h>
using namespace bitsquid;

namespace {
	Game &make_game(Allocator &a, Application &app)
	{
		return *a.make_new<MyGame>(app);
	}

	void set_opt(Vector<DynamicString>& args, const char* flag) {
		args.push_back(DynamicString(args.allocator(), flag));
	}

	void set_opt(Vector<DynamicString>& args, const char* name, const char* value) {
		set_opt(args, name); set_opt(args, value);
	}
}

Application* app = 0;
ApplicationOptions* options = 0;

extern "C" void bstech_android_entry(ANativeActivity* activity, void* savedState, size_t state_sz) {

	if (!memory_globals::is_inited()) {
		memory_globals::init();
	}

	Allocator& alloc = memory_globals::default_allocator();
	if (options) {
		alloc.make_delete(options);
		command_line::shutdown();
	}

	Vector<DynamicString> args(alloc);
	command_line::setup(activity);
	command_line::get(args);

	if ( args.size() < 2 ) {
		// set_opt(args, "-@1", "172.16.0.117");
		// set_opt(args, "-host", "@1");
		// set_opt(args, "-relay-host", "@1");
		// set_opt(args, "-wait", "2");
		// set_opt(args, "-project", "very-empty");
		//set_opt(args, "-editor-ini", "core/editor_slave/level_editor/settings");
		//set_opt(args, "-data-dir", "/Removable/SD/Android/data/com.bitsquid.BitSquidTech/files");
		//set_opt(args, "-data-dir", "/Removable/MicroSD/Android/data/com.bitsquid.BitSquidTech/files");
		//set_opt(args, "-bundle-dir", "/mnt/sdcard2/Android/data/com.bitsquid.BitSquidTech/files");
		//set_opt(args, "-apk-assets", "true");
		//set_opt(args, "-apk-ext-assets", "true");
		//set_opt(args, "-bundle-dir", activity->externalDataPath);
		//set_opt(args, "-data-dir", activity->externalDataPath);
	}

	logging::info("external-data-dir", activity->externalDataPath);
	logging::info("internal-data-dir", activity->internalDataPath);

	for ( int i = 0; i < args.size(); i++ ) {
		logging::info("CommandLine", "arg%d: %s", i, args[i].c_str());
	}
	
	options = alloc.make_new<ApplicationOptions>(args);
	options->set_internal_data_path(activity->internalDataPath);
	options->set_external_data_path(activity->externalDataPath);
		
	#ifdef DEVELOPMENT
		error_internal::set_console_wait_time(options->console_wait_time_after_error());
	#endif

	// Seed the global RNG.
	global_random().set_seed(random_seed());
	
	// This event is only ever triggered when the application is correctly destructed
	if (!Application::application_cleaned_up().wait(2000)) {
		logging::warning("Android", "restarting engine.");
		ANativeActivity_finish(activity);

		if (!Application::application_cleaned_up().wait(3000))
			XERROR("The engine did not shutdown properly. Exiting.");
	}

	apk_asset_manager::shutdown();
	apk_ext_manager::shutdown();
	app = memory_globals::application_static_allocator().make_new<Application>(*options, *make_game);
	app->init_activity(activity, savedState, state_sz);

	if (options->apk_assets())
		apk_asset_manager::init(activity->assetManager);

	if (options->apk_ext_assets())
		apk_ext_manager::init(options->obb_path());
	
	app->run();
}

/*
#include <foundation/platform.h>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

// Use a hash to store page_number -> page_size lookup
const unsigned PAGE_SIZE = 4096;
const unsigned MAX_PAGES = 1024 * 1024 * 1024 / PAGE_SIZE;
const unsigned MAX_ALLOC = ((1 << 8) - 1) * PAGE_SIZE;

// Since we don't insert allocations which are just one page into the table we know
// that it will contain at most MAX_PAGES/2 entries. 2*MAX_PAGES / 3 thus gives a
// load of 75 % at most.
const unsigned HASH_SIZE = (2*MAX_PAGES/3);
struct HashEntry {
	// Page number: p / PAGE_SIZE
	unsigned page : 24;
	
	// Number of pages allocated size: / PAGE_SIZE
	unsigned count : 8;
};
HashEntry page_hash[HASH_SIZE];

// Stored in hash table to indicate empty slot and stop of linear probing.
const unsigned FREE_STOP = 0;

// Stored in hash table to indicate empty slot, but continue linear probing.
const unsigned FREE_CONTINUE = 1;

bool inited = false;

void init()
{
	for (int i=0; i<HASH_SIZE; ++i) {
		page_hash[i].page = 0;
		page_hash[i].count = FREE_STOP;
	}
	inited = true;
}

void remember_size(void *p, size_t size)
{
	// Allocations with page size 1 is assumed to be the default, so we don't store those.
	size_t count = size / PAGE_SIZE;
	if (count == 1)
		return;
		
	if (!inited)
		init();
		
	size_t u = (char *)p - (char *)0;
	size_t page = u / PAGE_SIZE;
	
	// Linear probe for free page and store
	size_t h = page % HASH_SIZE;
	while (page_hash[h].count > FREE_CONTINUE)
		h = (h + 1) % HASH_SIZE;
	page_hash[h].page = page;
	page_hash[h].count = count;
}

void forget_size(void *p)
{
	size_t u = (char *)p - (char *)0;
	size_t page = u / PAGE_SIZE;
	
	// Probe for page
	size_t h = page % HASH_SIZE;
	while (page_hash[h].page != page) {
		if (page_hash[h].count == FREE_STOP)
			return;
		h = (h+1) % HASH_SIZE;
	}
	
	// Delete it
	page_hash[h].page = 0;
	page_hash[h].count = FREE_CONTINUE;
	
	// If next slot is FREE_STOP, convert all preceding FREE_CONTINUEs to
	// FREE_STOPs
	if (page_hash[ (h+1)%HASH_SIZE ].count == FREE_STOP) {
		while (page_hash[h].count == FREE_CONTINUE) {
			page_hash[h].count = FREE_STOP;
			h = (h + HASH_SIZE - 1) % HASH_SIZE;
		}
	}
}

size_t allocated_size(void *p)
{
	size_t u = (char *)p - (char *)0;
	size_t page = u / PAGE_SIZE;
	size_t h = page % HASH_SIZE;
	while (page_hash[h].page != page) {
		if (page_hash[h].count == FREE_STOP)
			return PAGE_SIZE;
		h = (h+1) % HASH_SIZE;
	}
	return page_hash[h].count * PAGE_SIZE;
}

bool contains(void* const* vals, unsigned size, void* value) {
	for (unsigned i = 0; i != size; ++i) {
		if (vals[i] == value)
			return true;
	}
	return false;
}

struct ANativeActivity;

extern "C" void bstech_android_entry(ANativeActivity* activity, void* savedState, size_t state_sz) {
	srand(time(0));
	LARGE_INTEGER freq, start, stop;
	QueryPerformanceFrequency(&freq);

	const unsigned ROUNDS = 100000;
	unsigned failed = 0, one_page = 0;

	void* picked[ROUNDS] = {0};
	unsigned n_picked = 0;

	double total_time = 0.0;

	for (unsigned i = 0; i != ROUNDS; ++i) {
		
		void* addr = (void*)((rand() << 16 | rand()) & 0xfffff000 | 0x68000000);
		unsigned size = (rand() % 15 + 1) * PAGE_SIZE;

		if (contains(&picked[0], n_picked, addr))
			continue;
		else
			picked[n_picked++] = addr;
		
		if (unsigned(addr) % PAGE_SIZE != 0)
			__android_log_print(ANDROID_LOG_INFO, "bstech", "invalid addr %p\n", addr);

		if (size == PAGE_SIZE)
			++one_page;

		remember_size(addr, size);

		QueryPerformanceCounter(&start);
		unsigned remembered_size = allocated_size(addr);
		QueryPerformanceCounter(&stop);
		total_time += ((stop.QuadPart - start.QuadPart) / (double)freq.QuadPart) * 1000000.0;

		if (size != remembered_size) {
			__android_log_print(ANDROID_LOG_INFO, "bstech", "allocated_size(%p) returned %d(%d) when %d(%d) was expected.\nn_picked = %d", addr, remembered_size,
				remembered_size / PAGE_SIZE, size, size / PAGE_SIZE, n_picked);

			++failed;
		}
	}
	
	__android_log_print(ANDROID_LOG_INFO, "bstech", "num_failures = %d, ratio = %.4f, num_one_page_allocs = %d\nsizeof(HashEntry) = %d\nlookup_table_size = %.2f K\ntotal_time = %f%cs\navg_time = %f%cs",
		failed, failed / (float)n_picked, one_page,
		sizeof(HashEntry), sizeof(page_hash) / 1024.f,
		total_time, 230,
		total_time / n_picked, 230);
}*/
